/*
HOPV -- Hybridized Orbital PreView
Copyright (C) 2006 Adomas Paltanavicius

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

$Id: hopv.c 17 2006-05-05 08:07:58Z admp $
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "zpr.h"

extern double drand48 (void);

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

int spacepoints;
struct Point {
	double x, y, z;
	double value;
	char neg;
} *points;

void setpoints(int n) {
	points = realloc(points, n * sizeof(points[0]));
	if (!points) {
		printf("Out of virtual memory!\n");
		exit(1);
	}
	spacepoints = n;
}

typedef double (*wave_fn_t)(int n, int Z, double r, double x, double y, double z);

/************************************************************
 * 1s                                                       *
 ***********************************************************/
double
wave_1s(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double)Z * r / (double) n;
	return 2. * sqrt(Z*Z*Z) * exp(-rho / 2.) * sqrt(1/(4*M_PI));
}

/************************************************************
 * 2s                                                       *
 ***********************************************************/
double
wave_2s(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return ((1/(2*M_SQRT2)) * (2. - rho) * sqrt(Z*Z*Z) * exp(-rho/2) *  //R
		    sqrt(1/(4*M_PI))); //Y
}

/************************************************************
 * 2p                                                       *
 ***********************************************************/
double
wave_2p_x(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return (1./(2.*sqrt(6.))) * rho * sqrt(Z*Z*Z) * exp(-rho/2) * //R
		    sqrt(3) * x/r * sqrt(1/(4*M_PI)); //Y
}

double
wave_2p_y(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return (1./(2.*sqrt(6.))) * rho * sqrt(Z*Z*Z) * exp(-rho/2) * //R
		    sqrt(3) * y/r * sqrt(1/(4*M_PI)); //Y
}

double
wave_2p_z(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return (1./(2.*sqrt(6.))) * rho * sqrt(Z*Z*Z) * exp(-rho/2) * //R
		    sqrt(3) * z/r * sqrt(1/(4*M_PI)); //Y
}

/************************************************************
 * sp                                                       *
 ***********************************************************/
double
waveh_sp_1(int n, int Z, double r, double x, double y, double z) {
	return sqrt(.5) * (wave_2s(n, Z, r, x, y, z) + wave_2p_z(n, Z, r, x, y, z));
}

double
waveh_sp_2(int n, int Z, double r, double x, double y, double z) {
	return sqrt(.5) * (wave_2s(n, Z, r, x,  y, z) - wave_2p_z(n, Z, r, x, y, z));
}

/************************************************************
 * sp2                                                      *
 ***********************************************************/
double
waveh_sp2_1(int n, int Z, double r, double x, double y, double z) {
	return 1./sqrt(3) * (wave_2s(n, Z, r, x, y, z) + 
			             M_SQRT2 * wave_2p_z(n, Z, r, x, y, z));
}

double
waveh_sp2_2(int n, int Z, double r, double x, double y, double z) {
	return 1./sqrt(3) * (wave_2s(n, Z, r, x, y, z) - 
			             M_SQRT2 * wave_2p_x(n, Z, r, x, y, z) + 
						 sqrt(3./2.) * wave_2p_y(n, Z, r, x, y, z));
}

double
waveh_sp2_3(int n, int Z, double r, double x, double y, double z) {
	return 1./sqrt(3) * (wave_2s(n, Z, r, x, y, z) - 
			             M_SQRT2 * wave_2p_x(n, Z, r, x, y, z) - 
					 	 sqrt(3./2.) * wave_2p_y(n, Z, r, x, y, z));
}

inline double absmax(double a, double b) {
	return fabs(a) > fabs(b) ? a : b;
}

/************************************************************
 * sp3                                                      *
 ***********************************************************/
double
waveh_sp3_1(int n, int Z, double r, double x, double y, double z) {
	return 1./2. * (wave_2s(n, Z, r, x, y, z) + 
			        wave_2p_x(n, Z, r, x, y, z) +
			        wave_2p_y(n, Z, r, x, y, z) +
			        wave_2p_z(n, Z, r, x, y, z));
}

double
waveh_sp3_2(int n, int Z, double r, double x, double y, double z) {
	return 1./2. * (wave_2s(n, Z, r, x, y, z) + 
			        wave_2p_x(n, Z, r, x, y, z) -
			        wave_2p_y(n, Z, r, x, y, z) -
			        wave_2p_z(n, Z, r, x, y, z));
}

double
waveh_sp3_3(int n, int Z, double r, double x, double y, double z) {
	return 1./2. * (wave_2s(n, Z, r, x, y, z) - 
			        wave_2p_x(n, Z, r, x, y, z) +
			        wave_2p_y(n, Z, r, x, y, z) -
			        wave_2p_z(n, Z, r, x, y, z));
}

double
waveh_sp3_4(int n, int Z, double r, double x, double y, double z) {
	return 1./2. * (wave_2s(n, Z, r, x, y, z) -
			        wave_2p_x(n, Z, r, x, y, z) -
			        wave_2p_y(n, Z, r, x, y, z) +
			        wave_2p_z(n, Z, r, x, y, z));
}


/************************************************************
 * 3s                                                       *
 ***********************************************************/
double
wave_3s(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return (1/(9*sqrt(3)) * (6 - 6*rho + rho*rho) * sqrt(Z*Z*Z) * exp(-rho/2) *
		    sqrt(1/(4*M_PI)));
}

/************************************************************
 * 3p                                                       *
 ***********************************************************/
double
wave_3p_x(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return (1/(9*sqrt(6)) * rho * (4 - rho) * sqrt(Z*Z*Z) * exp(-rho/2) *
			sqrt(3) * x/r * sqrt(1/(4*M_PI)));
}

double
wave_3p_y(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return (1/(9*sqrt(6)) * rho * (4 - rho) * sqrt(Z*Z*Z) * exp(-rho/2) *
			sqrt(3) * y/r * sqrt(1/(4*M_PI)));
}

double
wave_3p_z(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return (1/(9*sqrt(6)) * rho * (4 - rho) * sqrt(Z*Z*Z) * exp(-rho/2) *
			sqrt(3) * z/r * sqrt(1/(4*M_PI)));
}

/************************************************************
 * 3d                                                       *
 ***********************************************************/
static inline double
wave_3dR(int Z, double rho) {
	return	1/(9*sqrt(30)) * rho * rho * sqrt(Z*Z*Z) * exp(-rho/2);
}

double
wave_3d_xy(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return wave_3dR(Z, rho) * sqrt(60/4) * x * y / (r*r) * sqrt(1/(4*M_PI));
}

double
wave_3d_z2(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return wave_3dR(Z, rho) * sqrt(5./4.) * (2*z*z - (x*x - y*y))/(r*r) * sqrt(1/(4*M_PI));
}

/************************************************************
 * dsp3                                                     *
 ***********************************************************/
double
waveh_dsp3_eq1(int n, int Z, double r, double x, double y, double z) {
	return sqrt(.5) * (wave_3s(n, Z, r, x, y, z) + 
			           sqrt(2) * wave_3p_x(n, Z, r, x, y, z));
}

double
waveh_dsp3_eq2(int n, int Z, double r, double x, double y, double z) {
	return sqrt(.5) * (wave_3s(n, Z, r, x, y, z) -
			           1/sqrt(2) * wave_3p_x(n, Z, r, x, y, z) +
		               sqrt(3/2) * wave_3p_y(n, Z, r, x, y, z));
}

double
waveh_dsp3_ax1(int n, int Z, double r, double x, double y, double z) {
	return sqrt(.5) * (wave_3p_z(n, Z, r, x, y, z) + 
			           wave_3d_z2(n, Z, r, x, y, z));
}

/************************************************************
 * dsp3                                                     *
 ***********************************************************/
double
wave_4p_x(int n, int Z, double r, double x, double y, double z) {
	double rho = 2. * (double) Z * r / (double) n;
	return (1/(32*sqrt(15)) * rho * (20 - 10 * rho + rho*rho) *
			sqrt(Z*Z*Z) * exp(-rho/2) * sqrt(3) * x / r *
			sqrt(1/(4*M_PI)));
}

/************************************************************
 * Wave function list                                       *
 ***********************************************************/
struct {
	char *title;
	int n;
	double width;
	wave_fn_t fn;
} functions[] = {
	{ "1s",		1, 8., &wave_1s },
	{ "2s",		2, 25., &wave_2s },
	{ "2p_x",	2, 25, &wave_2p_x },
	{ "2p_y",   2, 25, &wave_2p_y },
	{ "2p_z",	2, 25, &wave_2p_z },
	{ "sp_1",	2, 25, &waveh_sp_1 },
	{ "sp_2",	2, 25, &waveh_sp_2 },
	{ "sp2_1",	2, 25, &waveh_sp2_1 },
	{ "sp2_2",	2, 25, &waveh_sp2_2 },
	{ "sp2_3",	2, 35, &waveh_sp2_3 },
	{ "sp3_1",	2, 35, &waveh_sp3_1 },
	{ "sp3_2",	2, 35, &waveh_sp3_2 },
	{ "sp3_3",	2, 35, &waveh_sp3_3 },
	{ "sp3_4",	2, 35, &waveh_sp3_4 },
	{ "3s",		3, 40, &wave_3s },
	{ "3p_x",	3, 40, &wave_3p_x },
	{ "3p_y",	3, 40, &wave_3p_y },
	{ "3p_z",	3, 40, &wave_3p_z },
	{ "3d_xy",	3, 40, &wave_3d_xy },
	{ "3d_zz",	3, 40, &wave_3d_z2 },
	{ "dsp3_eq1",3,55,&waveh_dsp3_eq1 },
	{ "dsp3_eq2",3,55,&waveh_dsp3_eq2 },
	{ "dsp3_ax1",3,55,&waveh_dsp3_ax1 },
	{ "4p_x",	4, 60,&wave_4p_x },
};
int opt_Z = 1;
int opt_n;
double opt_width;
enum { HALF_FULL = 0, HALF_X, HALF_Y, HALF_Z } opt_half;
wave_fn_t opt_fn;

void
randcoord(double *x, double *y, double *z) {
	*x = (double) drand48() * opt_width - opt_width/2;
	*y = (double) drand48() * opt_width - opt_width/2;
	*z = (double) drand48() * opt_width - opt_width/2;
}

inline void
euclidean_to_spherical(double x, double y, double z, 
                       double *rho, double *theta, double *phi) {
	if (rho)
		*rho = sqrt(x*x + y*y + z*z);
	if (theta)
		*theta = atan2(y, x);
	if (phi)
		*phi = atan2(sqrt(x*x + y*y), z);
}

void displayList(void);

void
compute(void) 
{
	double max = 1e-4;
	double x, y, z;
	int i;

	printf("Computing uniform points: ");
	
	/* As in Monte Carlo method. */
	for (i = 0; i < spacepoints; i++) {
		double rho;
		double val;
		randcoord(&x, &y, &z);
		if (opt_half &&
			((opt_half == HALF_X && x>0.) ||
			 (opt_half == HALF_Y && y>0.) ||
			 (opt_half == HALF_Z && z>0.))) {
			--i;
			continue;
		}
		euclidean_to_spherical(x, y, z, &rho, NULL, NULL);
		val = opt_fn(opt_n, opt_Z, rho, x, y, z);
		points[i].x = x;
		points[i].y = y;
		points[i].z = z;
		points[i].neg = val < 0;
		points[i].value = val*val;
		if (points[i].value <= .000016) {
			--i;
			continue;
		}
		if (i%1000 == 0) {
			glutPostRedisplay();
			printf(".");
			fflush(stdout);
		}
		if (points[i].value > max)
			max = points[i].value;
	}

	/* Scale. */
	for (i = 0; i < spacepoints; i++) {
		points[i].value /= max;
	}
	printf(" Done.\n");
	displayList();
}

inline int
colorize(struct Point *p, double c[3]) {
	c[3] = .9; /* alpha. */
	if (p->neg) {
		c[0] = 0.0;
		c[1] = pow(p->value, .7) * 2.6;
		if (c[1] > 1.)
			c[1] = 1.;
		c[2] = 0.0;
		return c[1] >= .2;
	} else {
		c[0] = pow(p->value, .7) * 2.6;
		if (c[0] > 1.)
			c[0] = 1.;
		c[1] = 0.0;
		c[2] = 0.0;
		return c[0] >= .2;
	}
}

int orbitalList;

void
displayList(void)
{
	double color[4];
	int i;
	
	printf("Creating graphical object: ");
	
	if (orbitalList)
		glDeleteLists(orbitalList, 1);
	orbitalList = glGenLists(1);
	glNewList(orbitalList, GL_COMPILE);

	glPushMatrix();
//	glLoadIdentity();
	glScaled(2/opt_width, 2/opt_width, 2/opt_width);
	glBegin(GL_POINTS);

	for (i = 0; i < spacepoints; i++) {
		if (colorize(&points[i], color)) {
			glColor4dv(color);
			glVertex3d(points[i].x, points[i].y, points[i].z);
		}
	}
	glEnd();
	glPopMatrix();
	glEndList();
	printf("Done.\n");
}

void
menuCallBack(int pos) {
	if (pos >= 0) {
		opt_n = functions[pos].n;
		opt_width = functions[pos].width;
		opt_fn = functions[pos].fn;
	} else {
		switch (pos) {
		case -1:
			exit(0);
		case -2:
			setpoints(spacepoints+25000);
			break;
		case -3:
			setpoints(spacepoints-25000);
			break;
		case -4:
			opt_half = HALF_X;
			break;
		case -5:
			opt_half = HALF_Y;
			break;
		case -6:
			opt_half = HALF_Z;
			break;
		case -7:
			opt_half = 0;
			break;
		case -8:
			setpoints(100000);
			break;
		case -9:
			setpoints(200000);
			break;
		case -10:
			setpoints(300000);
			break;
		case -11:
			setpoints(400000);
			break;
		case -12:
			setpoints(500000);
			break;
		case -13:
			setpoints(600000);
			break;
		case -14:
			setpoints(700000);
			break;
		}
	}
	compute();
	glutPostRedisplay();
}

void
setupMenu() {
	int menu = glutCreateMenu(menuCallBack);
	int optmenu = glutCreateMenu(menuCallBack);
	int i;
	
	glutSetMenu(menu);
	for (i = 0; i < sizeof(functions)/sizeof(functions[0]); i++) {
		glutAddMenuEntry(functions[i].title, i);
	}
	glutSetMenu(optmenu);
	glutAddMenuEntry("space+25000", -2);
	glutAddMenuEntry("space-25000", -3);
	glutAddMenuEntry("space=100000", -8);
	glutAddMenuEntry("space=200000", -9);
	glutAddMenuEntry("space=300000", -10);
	glutAddMenuEntry("space=400000", -11);
	glutAddMenuEntry("space=500000", -12);
	glutAddMenuEntry("space=600000", -13);
	glutAddMenuEntry("space=700000", -14);
	glutAddMenuEntry("1/2 x", -4);
	glutAddMenuEntry("1/2 y", -5);
	glutAddMenuEntry("1/2 z", -6);
	glutAddMenuEntry("2/2", -7);
	glutSetMenu(menu);
	glutAddSubMenu("Etc", optmenu);
	glutAddMenuEntry("X", -1);
	glutAttachMenu(GLUT_RIGHT_BUTTON);
}

void
display(void) {
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glCallList(orbitalList);
	glutSwapBuffers();
}

void
init(void)
{
	glShadeModel(GL_SMOOTH);
	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 


	GLfloat light_ambient[]  = { 0.0, 0.0, 0.0, 1.0 };
	GLfloat light_diffuse[]  = { 1.5, 1.5, 1.5, 1.5 };
	GLfloat light_specular[] = { .5, .5, .5, .5 };
	GLfloat light_position[] = { 0.0, 0.0, 0.0, 0.0 };

	GLfloat mat_ambient[]    = { 0.7, 0.7, 0.7, 1.0 };
	GLfloat mat_diffuse[]    = { 0.8, 0.8, 0.8, 1.0 };
	GLfloat mat_specular[]   = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat high_shininess[] = { 15.0 };

	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);

    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_FRONT, GL_SHININESS, high_shininess);
}

void
key(unsigned char k, int x, int y)
{
	switch (k) {
	case 'q':
	case 27:
		exit(0 + x-x + y-y); /* He he. */ 
		break;
	case 'f':
		glutFullScreen();
	}
	glutPostRedisplay();  
}

int
main(int argc, char **argv)
{
	printf("HPOV " __DATE__ " " __TIME__ "\n(C) Adomas Paltanavičius 2006\n\n");
	printf("HPOV comes with ABSOLUTELY NO WARRANTY; for details\n"
	       "see GNU General Public License.\n\n");
	
	glutInitWindowPosition(0, 0);
	glutInitWindowSize(700, 700);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutCreateWindow("HOPV: Hybridized Orbital Preview");
	glutDisplayFunc(display);
	glutKeyboardFunc(key);
	init();
	zprInit();
	opt_n = functions[0].n;
	opt_width = functions[0].width;
	opt_fn = functions[0].fn;
	setpoints(100000);
	compute();
	setupMenu();
	glutMainLoop();
	return 0;
}

/*
 * vi: ts=4 sw=4 ai
 */

